package jalview.bin.argparser;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.stream.Collectors;

import jalview.bin.Cache;
import jalview.util.ChannelProperties;

public enum Arg
{

  // Initialising arguments (BOOTSTRAP)
  HELP("h", "Display this help statement", Opt.UNARY, Opt.BOOTSTRAP),
  VERSION("v",
          "Display the version of "
                  + ChannelProperties.getProperty("app_name"),
          Opt.UNARY, Opt.BOOTSTRAP),
  HEADLESS(
          "Run Jalview in headless mode. No GUI interface will be created and Jalview will quit after all arguments have been processed.",
          Opt.UNARY, Opt.BOOTSTRAP),
  JABAWS("Set a different URL to connect to a JABAWS server.", Opt.STRING,
          Opt.BOOTSTRAP),
  NEWS("Show (or don't show) the news feed.", true, Opt.BOOLEAN,
          Opt.BOOTSTRAP),
  SPLASH("Show (or don't show) the About Jalview splash screen.", true,
          Opt.BOOLEAN, Opt.BOOTSTRAP),
  QUESTIONNAIRE(
          "Show (or don't show) the questionnaire if one is available.",
          true, Opt.BOOLEAN, Opt.BOOTSTRAP),
  USAGESTATS("Send (or don't send) initial launch usage stats.", true,
          Opt.BOOLEAN, Opt.BOOTSTRAP),
  WEBSERVICEDISCOVERY(
          "Attempt (or don't attempt) to connect to JABAWS web services.",
          true, Opt.BOOLEAN, Opt.BOOTSTRAP),
  PROPS("Use a file as the preferences file instead of the usual ~/"
          + ChannelProperties.getProperty("preferences.filename")
          + " file.", Opt.STRING, Opt.BOOTSTRAP),
  DEBUG("d", "Start Jalview in debug log level.", Opt.BOOLEAN,
          Opt.BOOTSTRAP),
  TRACE("Start Jalview in trace log level.", Opt.BOOLEAN, Opt.BOOTSTRAP,
          Opt.SECRET),
  QUIET("q",
          "Stop all output to STDOUT (after the Java Virtual Machine has started). Use ‑‑quiet a second time to stop all output to STDERR.",
          Opt.UNARY, Opt.MULTI, Opt.BOOTSTRAP),
  INITSUBSTITUTIONS(
          "Set ‑‑substitutions to be initially enabled (or initially disabled).",
          true, Opt.BOOLEAN, Opt.BOOTSTRAP, Opt.NOACTION),

  // Opening an alignment
  OPEN("Opens one or more alignment files or URLs in new alignment windows.",
          Opt.STRING, Opt.LINKED, Opt.INCREMENTDEFAULTCOUNTER, Opt.MULTI,
          Opt.GLOB, Opt.ALLOWSUBSTITUTIONS, Opt.INPUT, Opt.STORED),
  APPEND("Appends one or more alignment files or URLs to the open alignment window (or opens a new alignment if none already open).",
          Opt.STRING, Opt.LINKED, Opt.MULTI, Opt.GLOB,
          Opt.ALLOWSUBSTITUTIONS, Opt.INPUT),
  TITLE("Specifies the title for the open alignment window as string.",
          Opt.STRING, Opt.LINKED),
  COLOUR("color", // being a bit soft on the Americans!
          "Applies the colour scheme to the open alignment window. Valid values are:\n"
                  + "clustal,\n" + "blosum62,\n" + "pc-identity,\n"
                  + "zappo,\n" + "taylor,\n" + "gecos-flower,\n"
                  + "gecos-blossom,\n" + "gecos-sunset,\n"
                  + "gecos-ocean,\n" + "hydrophobic,\n"
                  + "helix-propensity,\n" + "strand-propensity,\n"
                  + "turn-propensity,\n" + "buried-index,\n"
                  + "nucleotide,\n" + "nucleotide-ambiguity,\n"
                  + "purine-pyrimidine,\n" + "rna-helices,\n"
                  + "t-coffee-scores,\n" + "sequence-id.",
          Opt.STRING, Opt.LINKED, Opt.ALLOWALL),
  FEATURES("Add a feature file or URL to the open alignment.", Opt.STRING,
          Opt.LINKED, Opt.MULTI, Opt.ALLOWSUBSTITUTIONS),
  TREE("Add a tree file or URL to the open alignment.", Opt.STRING,
          Opt.LINKED, Opt.MULTI, Opt.ALLOWSUBSTITUTIONS),
  SORTBYTREE(
          "Enforces sorting (or not sorting) the open alignment in the order of an attached phylogenetic tree.",
          true, Opt.LINKED, Opt.BOOLEAN, Opt.ALLOWALL),
  ANNOTATIONS("Add an annotations file or URL to the open alignment.",
          Opt.STRING, Opt.LINKED, Opt.MULTI, Opt.ALLOWSUBSTITUTIONS),
  SHOWANNOTATIONS(
          "Enforces showing (or not showing) alignment annotations.",
          Opt.BOOLEAN, Opt.LINKED, Opt.ALLOWALL),
  WRAP("Enforces wrapped (or not wrapped) alignment formatting.",
          Opt.BOOLEAN, Opt.LINKED, Opt.ALLOWALL),
  NOSTRUCTURE(
          "Do not open or process any 3D structure in the ‑‑open or ‑‑append files.",
          Opt.UNARY, Opt.LINKED, Opt.ALLOWALL),

  // Adding a 3D structure
  STRUCTURE(
          "Load a structure file or URL associated with a sequence in the open alignment.\n"
                  + "The sequence to be associated with can be specified with a following --seqid argument, or the subval modifier seqid=ID can be used. A subval INDEX can also be used to specify the INDEX-th sequence in the open alignment.",
          Opt.STRING, Opt.LINKED, Opt.MULTI, Opt.ALLOWSUBSTITUTIONS),
  SEQID("Specify the sequence name for the preceding --structure to be associated with.",
          Opt.STRING, Opt.LINKED, Opt.MULTI, Opt.ALLOWSUBSTITUTIONS),
  PAEMATRIX("Add a PAE json matrix file to the preceding --structure.",
          Opt.STRING, Opt.LINKED, Opt.MULTI, Opt.ALLOWSUBSTITUTIONS),
  TEMPFAC("Set the type of temperature factor. Possible values are:\n"
          + "default,\n" + "plddt.", Opt.STRING, Opt.LINKED),
  STRUCTUREVIEWER(
          "Set the structure viewer to use to open the 3d structure file specified in previous --structure to name. Possible values of name are:\n"
                  + "none,\n" + "jmol,\n" + "chimera,\n" + "chimerax,\n"
                  + "pymol.",
          Opt.STRING, Opt.LINKED, Opt.MULTI),
  NOTEMPFAC(
          "Do not show the temperature factor annotation for the preceding --structure.",
          Opt.UNARY, Opt.LINKED, Opt.ALLOWALL, Opt.SECRET), // keep this secret
                                                            // until it
  // works!
  SHOWSSANNOTATIONS(null, Opt.BOOLEAN, Opt.LINKED, Opt.ALLOWALL),

  // Outputting files
  IMAGE("Output an image of the open alignment window. Format is specified by the subval modifier, a following --type argument or guessed from the file extension. Valid formats/extensions are:\n"
          + "svg,\n" + "png,\n" + "eps,\n" + "html,\n" + "biojs.",
          Opt.STRING, Opt.LINKED, Opt.ALLOWSUBSTITUTIONS, Opt.ALLOWALL,
          Opt.REQUIREINPUT, Opt.OUTPUT),
  TYPE("Set the image format for the preceding --image to name. Valid values for name are: svg,\n"
          + "png,\n" + "eps,\n" + "html,\n" + "biojs.", Opt.STRING,
          Opt.LINKED, Opt.ALLOWALL),
  TEXTRENDERER(
          "Sets whether text in a vector image format (SVG, HTML, EPS) should be rendered as text or vector line-art. Possible values for name are:\n"
                  + "text,\n" + "lineart.",
          Opt.STRING, Opt.LINKED, Opt.ALLOWALL),
  SCALE("Sets a scaling for bitmap image format (PNG). Should be given as a floating point number. If used in conjunction with --width and --height then the smallest scaling will be used (scale, width and height provide bounds for the image).",
          Opt.STRING, Opt.LINKED, Opt.ALLOWALL),
  WIDTH("Sets a width for bitmap image format (PNG) with the height maintaining the aspect ratio. Should be given as a positive integer. If used in conjunction with --scale and --height then the smallest scaling will be used (scale, width and height provide bounds for the image).",
          Opt.STRING, Opt.LINKED, Opt.ALLOWALL),
  HEIGHT("Sets a height for bitmap image format (PNG) with the width maintaining the aspect ratio. Should be given as a positive integer. If used in conjunction with --scale and --width then the smallest scaling will be used (scale, width and height provide bounds for the image).",
          Opt.STRING, Opt.LINKED, Opt.ALLOWALL),
  OUTPUT("Export the open alignment to file filename. The format name is specified by the subval modifier format=name, a following --format name argument or guessed from the file extension. Valid format names (and file extensions) are:\n"
          + "fasta (fa, fasta, mfa, fastq),\n" + "pfam (pfam),\n"
          + "stockholm (sto, stk),\n" + "pir (pir),\n" + "blc (blc),\n"
          + "amsa (amsa),\n" + "json (json),\n" + "pileup (pileup),\n"
          + "msf (msf),\n" + "clustal (aln),\n" + "phylip (phy),\n"
          + "jalview (jvp, jar).", Opt.STRING, Opt.LINKED,
          Opt.ALLOWSUBSTITUTIONS, Opt.ALLOWALL, Opt.REQUIREINPUT,
          Opt.OUTPUT),
  FORMAT("Sets the format for the preceding --output file. Valid formats are:\n"
          + "fasta,\n" + "pfam,\n" + "stockholm,\n" + "pir,\n" + "blc,\n"
          + "amsa,\n" + "json,\n" + "pileup,\n" + "msf,\n" + "clustal,\n"
          + "phylip,\n" + "jalview.", Opt.STRING, Opt.LINKED, Opt.ALLOWALL),
  GROOVY("Process a groovy script in the file for the open alignment.",
          Opt.STRING, Opt.LINKED, Opt.MULTI, Opt.ALLOWSUBSTITUTIONS,
          Opt.ALLOWALL),
  BACKUPS("Enable (or disable) writing backup files when saving an ‑‑output file. This applies to the current open alignment.  To apply to all ‑‑output and ‑‑image files, use after ‑‑all.",
          true, Opt.BOOLEAN, Opt.LINKED, Opt.ALLOWALL),
  OVERWRITE(
          "Enable (or disable) overwriting of output files without backups enabled. This applies to the current open alignment.  To apply to all ‑‑output and ‑‑image files, use after ‑‑all.",
          Opt.BOOLEAN, Opt.LINKED, Opt.ALLOWALL),
  CLOSE("Close the current open alignment window. This occurs after other output arguments. This applies to the current open alignment.  To apply to all ‑‑output and ‑‑image files, use after ‑‑all.",
          Opt.UNARY, Opt.LINKED, Opt.ALLOWALL),

  // controlling flow of arguments
  NEW("Move on to a new alignment window. This will ensure --append will start a new alignment window and other linked arguments will apply to the new alignment window.",
          Opt.UNARY, Opt.MULTI, Opt.NOACTION, Opt.INCREMENTDEFAULTCOUNTER),
  SUBSTITUTIONS(
          "The following argument values allow (or don't allow) subsituting filename parts. This is initially true. Valid substitutions are {basename} - the filename-without-extension of the currently --opened file (or first --appended file),\n"
                  + "{dirname}, - the directory (folder) name of the currently --opened file (or first --appended file),\n"
                  + "{argfilebasename} - the filename-without-extension of the current --argfile,\n"
                  + "{argfiledirname} - the directory (folder) name of the current --argfile,\n"
                  + "{n} - the value of the index counter (starting at 0).\n"
                  + "{++n} - increase and substitute the value of the index counter,\n"
                  + "{} - the value of the current alignment window default index.",
          true, Opt.BOOLEAN, Opt.MULTI, Opt.NOACTION),
  ARGFILE("Open one or more files filename and read, line-by-line, as arguments to Jalview.\n"
          + "Values in an argfile should be given with an equals sign (\"=\") separator with no spaces.\n"
          + "Note that if you use one or more --argfile arguments then all other non-initialising arguments will be ignored.",
          Opt.STRING, Opt.MULTI, Opt.BOOTSTRAP, Opt.GLOB,
          Opt.ALLOWSUBSTITUTIONS),
  NPP("n++",
          "Increase the index counter used in argument value substitutions.",
          Opt.UNARY, Opt.MULTI, Opt.NOACTION),
  ALL("Apply the following output arguments to all sets of linked arguments.",
          Opt.BOOLEAN, Opt.MULTI, Opt.NOACTION),
  OPENED("Apply the following output arguments to all of the last --open'ed set of linked arguments.",
          Opt.BOOLEAN, Opt.MULTI, Opt.NOACTION),
  QUIT("After all files have been opened, appended and output, quit Jalview. In ‑‑headless mode this already happens.",
          Opt.UNARY),

  // secret options
  TESTOUTPUT(
          "Allow specific stdout information.  For testing purposes only.",
          Opt.UNARY, Opt.BOOTSTRAP, Opt.SECRET), // do not show this to the user
  SETPROP("Set an individual Java System property.", Opt.STRING, Opt.MULTI,
          Opt.BOOTSTRAP, Opt.SECRET), // not in use yet
  NIL("This argument does nothing on its own, but can be used with linkedIds.",
          Opt.UNARY, Opt.LINKED, Opt.MULTI, Opt.NOACTION, Opt.SECRET),

  // private options (inserted during arg processing)
  SETARGFILE(
          "Sets the current value of the argfilename.  Inserted before argfilecontents.",
          Opt.UNARY, Opt.LINKED, Opt.STRING, Opt.MULTI, Opt.PRIVATE,
          Opt.NOACTION),
  UNSETARGFILE(
          "Unsets the current value of the argfilename.  Inserted after argfile contents.",
          Opt.UNARY, Opt.LINKED, Opt.MULTI, Opt.PRIVATE, Opt.NOACTION),

  // these last two have no purpose in the normal Jalview application but are
  // used by jalview.bin.Launcher to set memory settings. They are not used by
  // argparser but are here for Usage statement reasons.
  JVMMEMPC(
          "Limit maximum heap size (memory) to PERCENT% of total physical memory detected. This defaults to 90 if total physical memory can be detected.\n"
                  + "The equals sign (\"=\") separator must be used with no spaces.",
          Opt.NOACTION, Opt.BOOTSTRAP, Opt.STRING),
  JVMMEMMAX(
          "Limit maximum heap size (memory) to MAXMEMORY. MAXMEMORY can be specified in bytes, kilobytes(k), megabytes(m), gigabytes(g) or if you're lucky enough, terabytes(t). This defaults to 32g if total physical memory can be detected, or to 8g if total physical memory cannot be detected.\n"
                  + "The equals sign (\"=\") separator must be used with no spaces.",
          Opt.NOACTION, Opt.BOOTSTRAP, Opt.STRING),

  ;

  public static enum Opt
  {
    BOOLEAN, // This Arg can be specified as --arg or --noarg to give true or
             // false. A default can be given with setOptions(bool, Opt....).
             // Use ArgParser.isSet(Arg) to see if this arg was not specified.
    STRING, // This Arg can accept a value either through --arg=value or --arg
            // value.
    UNARY, // This Arg is a boolean value, true if present, false if not. Like
           // BOOLEAN but without the --noarg option.
    MULTI, // This Arg can be specified multiple times. Multiple values are
           // stored in the ArgValuesMap (along with their positional index) for
           // each linkedId.
    LINKED, // This Arg can be linked to others through a --arg[linkedId] or
            // --arg[linkedId]=value. If no linkedId is specified then the
            // current default linkedId will be used.
    NODUPLICATEVALUES, // This Arg can only have one value (per linkedId). The
                       // first value will be used and subsequent values ignored
                       // with a warning.
    BOOTSTRAP, // This Arg value(s) can be determined at an earlier stage than
               // non-BOOTSTRAP Args. Substitutions do not happen in BOOTSTRAP
               // Args and they cannot be linked or contain SubVals. See
               // jalview.bin.argparser.BootstrapArgs.
    GLOB, // This Arg can expand wildcard filename "globs" (e.g.
          // path/*/filename*). If the Arg value is given as --arg filename*
          // then the shell will have expanded the glob already, but if
          // specified as --arg=filename* then the Java glob expansion method
          // will be used (see FileUtils.getFilenamesFromGlob()). Note that this
          // might be different from the shell expansion rules.
    NOACTION, // This Arg does not perform a data task, usually used to control
              // flow in ArgParser.parse(args).
    ALLOWSUBSTITUTIONS, // This Arg allows substitutions in its linkedId,
                        // SubVals and values.
    PRIVATE, // This Arg is used internally, and cannot be specified by the
             // user.
    SECRET, // This Arg is used by development processes and although it can be
            // set by the user, it is not displayed to the user.
    ALLOWALL, // This Arg can use the '*' linkedId to apply to all known
              // linkedIds
    INCREMENTDEFAULTCOUNTER, // If an Arg has this option and the default
                             // linkedId is used, the defaultLinkedIdCounter is
                             // incremented *first*.
    INPUT, // This Arg counts as an input for REQUIREINPUT
    REQUIREINPUT, // This Arg can only be applied via --all if there is an
                  // input (i.e. --open or --append)
    OUTPUT, // This Arg provides an output filename. With Opt.ALLOWALL *.ext is
            // shorthand for --all --output={basename}.ext
    STORED, // This Arg resets and creates a new set of "opened" linkedIds
  }

  private final String[] argNames;

  private Opt[] argOptions;

  private boolean defaultBoolValue = false;

  private String description = null;

  private Arg(String description, Opt... options)
  {
    this(null, description, false, options);
  }

  private Arg(String description, boolean defaultBoolean, Opt... options)
  {
    this(null, description, defaultBoolean, options);
  }

  private Arg(String alternativeName, String description, Opt... options)
  {
    this(alternativeName, description, false, options);
  }

  private Arg(String alternativeName, String description,
          boolean defaultBoolean, Opt... options)
  {
    this.argNames = alternativeName != null
            ? new String[]
            { this.getName(), alternativeName }
            : new String[]
            { this.getName() };
    this.description = description;
    this.defaultBoolValue = defaultBoolean;
    this.setOptions(options);
  }

  public String argString()
  {
    return argString(false);
  }

  public String negateArgString()
  {
    return argString(true);
  }

  private String argString(boolean negate)
  {
    StringBuilder sb = new StringBuilder(ArgParser.DOUBLEDASH);
    if (negate && hasOption(Opt.BOOLEAN))
      sb.append(ArgParser.NEGATESTRING);
    sb.append(getName());
    return sb.toString();
  }

  public String toLongString()
  {
    StringBuilder sb = new StringBuilder();
    sb.append(this.getClass().getName()).append('.').append(this.name());
    sb.append('(');
    if (getNames().length > 0)
      sb.append('"');
    sb.append(String.join("\", \"", getNames()));
    if (getNames().length > 0)
      sb.append('"');
    sb.append(")\n");
    sb.append("\nOpt: ");
    // map List<Opt> to List<String> for the String.join
    List<String> optList = Arrays.asList(argOptions).stream()
            .map(opt -> opt.name()).collect(Collectors.toList());
    sb.append(String.join(", ", optList));
    sb.append("\n");
    return sb.toString();
  }

  public String[] getNames()
  {
    return argNames;
  }

  public String getName()
  {
    return this.name().toLowerCase(Locale.ROOT).replace('_', '-');
  }

  @Override
  public final String toString()
  {
    return getName();
  }

  public boolean hasOption(Opt o)
  {
    if (argOptions == null)
      return false;
    for (Opt option : argOptions)
    {
      if (o == option)
        return true;
    }
    return false;
  }

  protected void setOptions(Opt... options)
  {
    this.argOptions = options;
  }

  protected boolean getDefaultBoolValue()
  {
    return defaultBoolValue;
  }

  protected String getDescription()
  {
    return description;
  }

  public static String booleanArgString(Arg a)
  {
    StringBuilder sb = new StringBuilder(a.argString());
    if (a.hasOption(Opt.BOOLEAN))
    {
      sb.append('/');
      sb.append(a.negateArgString());
    }
    return sb.toString();
  }

  public static final String usage()
  {
    StringBuilder sb = new StringBuilder();

    sb.append(ChannelProperties.getProperty("app_name"));
    String version = Cache.getDefault("VERSION", null);
    if (version != null)
    {
      sb.append(" version ");
      sb.append(Cache.getDefault("VERSION", "unknown"));
    }
    sb.append(System.lineSeparator());
    sb.append("Usage: jalview [files...] [args]");
    sb.append(System.lineSeparator());
    sb.append(System.lineSeparator());

    int maxArgLength = 0;
    for (Arg a : EnumSet.allOf(Arg.class))
    {
      if (a.hasOption(Opt.PRIVATE) || a.hasOption(Opt.SECRET))
        continue;

      StringBuilder argSb = new StringBuilder();
      argSb.append(a.hasOption(Opt.BOOLEAN) ? booleanArgString(a)
              : a.argString());
      if (a.hasOption(Opt.STRING))
        argSb.append("=value");
      if (argSb.length() > maxArgLength)
        maxArgLength = argSb.length();
    }

    // might want to sort these
    for (Arg a : EnumSet.allOf(Arg.class))
    {
      if (a.hasOption(Opt.PRIVATE) || a.hasOption(Opt.SECRET))
        continue;
      StringBuilder argSb = new StringBuilder();
      argSb.append(a.hasOption(Opt.BOOLEAN) ? booleanArgString(a)
              : a.argString());
      if (a.hasOption(Opt.STRING))
        argSb.append("=value");
      Iterator<String> descLines = null;
      if (a.getDescription() != null)
      {
        descLines = Arrays.stream(a.getDescription().split("\\n"))
                .iterator();
      }
      sb.append(String.format("%-" + maxArgLength + "s", argSb.toString()));
      boolean first = true;
      if (descLines != null)
      {
        while (descLines.hasNext())
        {
          if (first)
            sb.append(" - ");
          else
            sb.append(" ".repeat(maxArgLength + 3));
          sb.append(descLines.next());
          sb.append(System.lineSeparator());
          first = false;
        }
      }

      List<String> options = new ArrayList<>();

      if (a.hasOption(Opt.BOOLEAN))
      {
        options.add("default " + (a.getDefaultBoolValue() ? a.argString()
                : a.negateArgString()));
      }

      if (a.hasOption(Opt.MULTI))
      {
        options.add("multiple");
      }

      if (a.hasOption(Opt.LINKED))
      {
        options.add("can be linked");
      }

      if (a.hasOption(Opt.GLOB))
      {
        options.add("allows file globs");
      }

      if (a.hasOption(Opt.ALLOWSUBSTITUTIONS))
      {
        options.add("allows substitutions");
      }

      if (a.hasOption(Opt.ALLOWALL))
      {
        options.add("can be applied to all linked arguments");
      }

      if (a.hasOption(Opt.PRIVATE))
      {
        options.add("for internal use only");
      }

      if (a.hasOption(Opt.SECRET))
      {
        options.add("for development use only");
      }

      if (options.size() > 0)
      {
        if (first)
          sb.append(" - ");
        else
          sb.append(" ".repeat(maxArgLength + 3));
        sb.append("(");
        sb.append(String.join("; ", options));
        sb.append(')');
        sb.append(System.lineSeparator());
      }
      sb.append(System.lineSeparator());
    }
    return sb.toString();
  }
}